import base64
import bz2
import json
from datetime import datetime
from enum import Enum
from typing import Any
from typing import Optional, Dict, List

from pydantic import BaseModel
from pydantic import validator

from switcore.ui.button import Button
from switcore.ui.collection_entry import CollectionEntry
from switcore.ui.container import Container
from switcore.ui.divider import Divider
from switcore.ui.file import File
from switcore.ui.header import Header
from switcore.ui.html_frame import HtmlFrame
from switcore.ui.image import Image
from switcore.ui.input import Input
from switcore.ui.select import Select
from switcore.ui.signIn_page import SignInPage
from switcore.ui.tabs import Tabs
from switcore.ui.text_paragraph import TextParagraph
from switcore.ui.textarea import Textarea
from switcore.ui.datepicker import DatePicker
from switcore.ui.info_card import InfoCard


class UserInfo(BaseModel):
    user_id: str
    organization_id: str


class UserPreferences(BaseModel):
    language: str
    time_zone_offset: str
    color_theme: str


class Resource(BaseModel):
    resource_type: str
    id: str
    created_at: datetime
    edited_at: datetime
    content: str
    content_formatted: Dict[str, Any]
    attachments: Optional[List[Dict[str, Any]]]
    files: Optional[List[Dict[str, Any]]]
    creator: Dict[str, Any]


class UserActionType(str, Enum):
    right_panel_open = "right_panel_open"
    presence_sync = "presence_sync"
    user_commands_chat = "user_commands.extensions:chat"
    user_commands_chat_extension = "user_commands.chat_extension"
    user_commands_chat_commenting = "user_commands.extensions:chat_commenting"
    user_commands_context_menus_message = "user_commands.context_menus:message"
    user_commands_context_menus_message_comment = "user_commands.context_menus:message_comment"
    view_actions_drop = "view_actions.drop"
    view_actions_input = "view_actions.input"
    view_actions_submit = "view_actions.submit"
    view_actions_oauth_complete = "view_actions.oauth_complete"


class UserAction(BaseModel):
    type: UserActionType
    id: str
    slash_command: str
    resource: Resource | None
    value: str | None = None


class Context(BaseModel):
    workspace_id: str
    channel_id: str


ElementType = (CollectionEntry | Button | Divider | File | HtmlFrame | Input | Select
               | SignInPage | TextParagraph | Image | Textarea | Container | Tabs | DatePicker | InfoCard)


def contain_only_dict(elements_data: list[dict | ElementType]) -> bool:
    for element_data in elements_data:
        if isinstance(element_data, ElementType):
            return False

    return True


def get_element_type(element_data: dict):
    element_type_str = element_data.get('type')
    if element_type_str == 'collection_entry':
        return CollectionEntry
    elif element_type_str == 'button':
        return Button
    elif element_type_str == 'divider':
        return Divider
    elif element_type_str == 'file':
        return File
    elif element_type_str == 'html_frame':
        return HtmlFrame
    elif element_type_str == 'text_input':
        return Input
    elif element_type_str == 'select':
        return Select
    elif element_type_str == 'sign_in_page':
        return SignInPage
    elif element_type_str == 'text':
        return TextParagraph
    elif element_type_str == 'textarea':
        return Textarea
    elif element_type_str == 'image':
        return Image
    elif element_type_str == 'container':
        return Container
    elif element_type_str == 'tabs':
        return Tabs
    elif element_type_str == 'datepicker':
        return DatePicker
    elif element_type_str == 'info_card':
        return InfoCard
    else:
        raise ValueError(f"Unknown element type: {element_type_str}")


class Body(BaseModel):
    elements: list[ElementType]

    class Config:
        smart_union = True

    def __init__(self, **data: Any) -> None:
        elements_data = data.get('elements', [])

        if contain_only_dict(elements_data):
            _elements: list[ElementType] = []
            for element_data in elements_data:
                element_type = get_element_type(element_data)
                element = element_type(**element_data)
                _elements.append(element)
            super().__init__(elements=_elements)
        else:
            super().__init__(**data)

    def get_element_by_action_id(self, action_id: str) -> ElementType:
        """
        Get element by action_id if it exists, otherwise raise ValueError
        """
        for element in self.elements:
            if getattr(element, 'action_id', None) is None:
                continue

            if element.action_id == action_id:
                return element

        raise ValueError(f"Element with action_id: {action_id} not found")


class Footer(BaseModel):
    buttons: list[Button]


class ViewCallbackType(str, Enum):
    update = "views.update"
    initialize = "views.initialize"
    open = "views.open"
    push = "views.push"
    close = "views.close"


class AttachmentCallbackTypes(str, Enum):
    share_channel = "attachments.share.channel"


class AttachmentView(BaseModel):
    state: str | bytes
    header: Header
    footer: Footer | None
    body: Body


class View(AttachmentView):
    view_id: str


class ViewResponse(BaseModel):
    callback_type: ViewCallbackType
    new_view: View


class AttachmentResponse(BaseModel):
    callback_type: AttachmentCallbackTypes
    attachments: AttachmentView


class PlatformTypes(str, Enum):
    DESKTOP = 'Desktop'
    IOS = 'iOS'
    ANDROID = 'Android'


class SwitRequest(BaseModel):
    platform: PlatformTypes
    time: datetime
    app_id: str
    user_info: UserInfo
    user_preferences: UserPreferences
    context: Context
    user_action: UserAction
    current_view: View | AttachmentView | None

    @validator('current_view', pre=True)
    def empty_dict_to_null(cls, v):
        if v == {}:
            return None
        return v


class SwitResponse(BaseModel):
    callback_type: ViewCallbackType | AttachmentCallbackTypes
    new_view: View | None
    attachments: AttachmentView | None
    reference_view_id: str | None


class BaseState(BaseModel):
    autoincrement_id: int = 1

    @classmethod
    def from_bytes(cls, byte: bytes):
        d = json.loads(bz2.decompress(base64.b64decode(byte)).decode("utf-8"))
        return cls(**d)

    def to_bytes(self) -> bytes:
        return base64.b64encode(bz2.compress(json.dumps(self.dict()).encode("utf-8"), 1))



import datetime
import os

from fastapi import FastAPI

from switcore.action.schemas import SwitRequest, PlatformTypes, UserInfo, UserPreferences, Context, UserAction, \
    UserActionType, View, Body
from switcore.ui.header import Header


def create_fastapi_app():
    os.environ["SWIT_CLIENT_ID"] = "test_client_id"
    os.environ["SWIT_CLIENT_SECRET"] = "test_client_secret"
    os.environ["APPS_ID"] = "test_apps_id"
    os.environ["DB_USERNAME"] = "test_db_username"
    os.environ["DB_PASSWORD"] = "test_db_password"
    os.environ["DB_NAME"] = "test_db_name"
    os.environ["ENV_OPERATION"] = "local"
    os.environ["BASE_URL"] = "test_base_url"
    os.environ["LOCALIZE_PROJECT_ID"] = "test_localize_project_id"
    os.environ["DB_HOST"] = "test_db_host"
    os.environ[
        "SCOPES"] = ("imap:write+imap:read+user:read+message:write+channel:read+workspace:read+project"
                     ":read+project:write+task:read+task:write")
    os.environ["BOT_REDIRECT_URL"] = "/auth/callback/bot"
    os.environ["USER_REDIRECT_URL"] = "/auth/callback/user"

    return FastAPI()


def create_swit_request(view_id: str, action_id: str):
    return SwitRequest(
        platform=PlatformTypes.DESKTOP,
        time=datetime.datetime.now(),
        app_id="test_app_id",
        user_info=UserInfo(
            user_id="test_user_id",
            organization_id="test_organization_id"
        ),
        user_preferences=UserPreferences(
            language="ko",
            time_zone_offset="+0900",
            color_theme="light"
        ),
        context=Context(
            workspace_id="test_workspace_id",
            channel_id="test_channel_id"
        ),
        user_action=UserAction(
            type=UserActionType.view_actions_submit,
            id=action_id,
            slash_command="test_slash_command",
        ),
        current_view=View(
            view_id=view_id,
            state="state",
            header=Header(
                title="test_title",
            ),
            body=Body(),
        ))



import json
import os

import httpx


async def set_translate():
    localise_token: str = os.getenv('LOCALIZE_PROJECT_ID', None)

    assert localise_token is not None, "LOCALIZE_PROJECT_ID is not set check .env file"

    params = {
        "project_id": localise_token,
        "tags": "webhook proxy",
    }

    async with httpx.AsyncClient() as client:
        res = await client.get("https://ur.swit.support/get_trans", params=params, timeout=60)

    if not res.is_success:
        return res.json()

    with open("./translate.json", "w", encoding="utf-8") as file:
        file.write(json.dumps(res.json(), indent=2))

    return res.json()


def get_translated_text(lang: str, key: str, number: str | None = None) -> str:
    with open("translate.json", "r", encoding="utf-8") as file:
        translate_info = json.load(file)

    appname: str = translate_info[lang]["webhookproxy.appname"]
    text: str = translate_info[lang][key]
    text = text.replace("{appName}", appname)
    if number:
        text = text.replace("{n}", number)
    return text



import json
import unittest

from switcore.ui.element_components import OpenLink
from switcore.ui.select import Select, Option, OptionGroup, SelectSearch


class SelectTest(unittest.TestCase):

    def test_valid_select01(self):
        select = Select(
            trigger_on_input=True,
            options=[
                Option(
                    label="test label1",
                    action_id="action_id1"
                ),
                Option(
                    label="test label2",
                    action_id="action_id2"
                ),
            ]
        )
        expected = {
            'type': 'select',
            'multiselect': False,
            'trigger_on_input': True,
            'options': [
                {
                    'label': 'test label1',
                    'action_id': 'action_id1'
                },
                {
                    'label': 'test label2',
                    'action_id': 'action_id2'
                }
            ],
            'option_groups': []
        }
        # print(json.dumps(expected, indent=4))
        self.assertEqual(expected, select.dict(exclude_none=True))

    def test_valid_select02(self):
        select = Select(
            trigger_on_input=True,
            options=[
                Option(
                    label="test label1",
                    action_id="action_id1",
                    static_action=OpenLink(
                        link_url="https://www.google.com"
                    )
                ),
                Option(
                    label="test label2",
                    action_id="action_id2"
                ),
            ]
        )
        expected = {
            'type': 'select',
            'multiselect': False,
            'options': [{'action_id': 'action_id1',
                         'label': 'test label1',
                         'static_action': {'link_url': 'https://www.google.com',
                                           'action_type': 'open_link'}},
                        {'action_id': 'action_id2', 'label': 'test label2'}],
            'option_groups': [],
            'trigger_on_input': True,
        }
        print(json.dumps(expected, indent=4))
        self.assertEqual(expected, select.dict(exclude_none=True))

    def test_valid_select03(self):
        select = Select(
            trigger_on_input=True,
            options=[
                Option(
                    label="test label1",
                    action_id="action_id1",
                    static_action=OpenLink(
                        link_url="https://www.google.com"
                    )
                ),
                Option(
                    label="test label2",
                    action_id="action_id2"
                ),
            ],
            option_groups=[
                OptionGroup(label="test group1", options=[Option(label="test label1", action_id="action_id1")]),
                OptionGroup(label="test group2", options=[Option(label="test label2", action_id="action_id2")]),
            ],
            search=SelectSearch(
                disabled=False,
                placeholder="test placeholder",
                value="test value",
                action_id="action_id"
            )
        )

        expected: dict = {
            'type': 'select',
            'multiselect': False,
            'options': [{'action_id': 'action_id1',
                         'label': 'test label1',
                         'static_action': {'link_url': 'https://www.google.com',
                                           'action_type': 'open_link'}},
                        {'action_id': 'action_id2', 'label': 'test label2'}],
            'option_groups': [{'label': 'test group1',
                               'options': [{'action_id': 'action_id1', 'label': 'test label1'}]},
                              {'label': 'test group2',
                               'options': [{'action_id': 'action_id2', 'label': 'test label2'}]}],
            'trigger_on_input': True,
            'search': {'action_id': 'action_id',
                       'disabled': False,
                       'placeholder': 'test placeholder',
                       'value': 'test value'},
        }
        self.assertEqual(expected, select.dict(exclude_none=True))



from enum import Enum

from pydantic import BaseModel


class SizeTypes(str, Enum):
    small = "small"
    medium = "medium"
    large = "large"


class Textarea(BaseModel):
    type: str = "textarea"
    action_id: str
    placeholder: str | None
    value: str | None
    height: SizeTypes
    disabled: bool = False



from pydantic import BaseModel

from switcore.ui.button import Button
from switcore.ui.select import Select
from switcore.ui.datepicker import DatePicker


class Container(BaseModel):
    type: str = "container"
    elements: list[Button | Select | DatePicker]



from enum import Enum

from pydantic import BaseModel

from switcore.ui.element_components import SubText, Tag, StaticAction
from switcore.ui.image import Image
from switcore.ui.text_paragraph import TextParagraph


class TextStyle(BaseModel):
    bold: bool = False
    color: str
    size: str
    max_lines: int


class BackgroundType(str, Enum):
    none = "none"
    lightblue = "lightblue"


class Background(BaseModel):
    color: BackgroundType = BackgroundType.none


class MetadataItem(BaseModel):
    type: str
    content: str | None
    style: dict | None
    image_url: str | None


class TextSection(BaseModel):
    text: TextParagraph
    metadata_items: list[SubText | Image | Tag] | None


class VerticalAlignmentTypes(str, Enum):
    top = "top"
    middle = "middle"
    bottom = "bottom"


class CollectionEntry(BaseModel):
    type: str = "collection_entry"
    start_section: Image | None
    text_sections: list[TextSection]
    vertical_alignment: VerticalAlignmentTypes = VerticalAlignmentTypes.top
    background: Background | None
    action_id: str | None
    static_action: StaticAction | None
    draggable: bool = False



from pydantic import BaseModel

from switcore.ui.Icon import Icon
from switcore.ui.button import Button


class IntegratedService(BaseModel):
    icon: Icon


class SignInPage(BaseModel):
    id: str | None = None
    type: str = "sign_in_page"
    integrated_service: IntegratedService
    title: str
    description: str
    button: Button



from collections import defaultdict
from urllib.parse import quote, unquote

from switcore.type import DrawerHandler


def escape(text: str) -> str:
    return quote(unquote(text), safe='')


class PathResolver:
    def __init__(self, action_id: str, paths: list[int | str] | None = None) -> None:
        if paths is None:
            paths = []

        self._escaped_action_id: str = escape(action_id)
        self._escaped_paths: list[str] = []
        for path in paths:
            self._add_path(path)

    def __str__(self):
        return self.combined_path

    @property
    def combined_path(self) -> str:
        """
        :return escaped action_id + escaped paths string
        """
        escaped: str = '/'.join(self._escaped_paths)
        return f'{self._escaped_action_id}/{escaped}'

    @property
    def id(self) -> str:
        return unquote(self._escaped_action_id)

    @property
    def key(self) -> str:
        """
        :return: escaped action_id
        """
        return self._escaped_action_id

    @property
    def paths(self) -> list[int | str]:
        ret: list[int | str] = []
        for path in self._escaped_paths:
            if path.isdigit():
                ret.append(int(path))
            else:
                ret.append(unquote(path))

        return ret

    @staticmethod
    def from_combined(combined: str) -> 'PathResolver':
        """
        :param combined: escaped action_id + escaped paths
        """
        arr: list[str] = combined.split('/')
        paths: list[str] = arr[1:]

        if len(paths) == 1 and paths[0] == '':
            paths = []

        return PathResolver(arr[0], paths)

    def _add_path(self, path: int | str):
        assert isinstance(path, int) or isinstance(path, str), "only int or str is allowed"

        self._escaped_paths.append(escape(str(path)))


class ActivityRouter:

    def __init__(self) -> None:
        self.handlers: dict[str, DrawerHandler] = dict()
        self.view_actions: dict[str, set[str]] = defaultdict(set)

    def register(self, action_id: str, view_ids: list[str] | None = None):
        def decorator(func):
            escaped_action_id = escape(action_id)
            self.handlers[escaped_action_id] = func

            if view_ids is None:
                self.view_actions[escaped_action_id].add('*')
            else:
                for view_id in view_ids:
                    self.view_actions[escaped_action_id].add(escape(view_id))

            return func

        return decorator



import unittest

from switcore.ui.datepicker import DatePicker


class DatePickerTest(unittest.TestCase):

    def test_valid(self):
        datepicker = DatePicker(
            action_id='test_action_id',
            placeholder='test_placeholder',
            trigger_on_input=True,
            value='2022-11-23T00:00:00.000Z')

        expected = {
            'type': 'datepicker',
            'action_id': 'test_action_id',
            'placeholder': 'test_placeholder',
            'trigger_on_input': True,
            'value': '2022-11-23T00:00:00.000Z'
        }

        self.assertEqual(expected, datepicker.dict(exclude_none=True))



from pydantic import BaseModel


class SwitToken(BaseModel):
    access_token: str
    expires_in: int
    refresh_token: str
    token_type: str
    scope: str


class Payload(BaseModel):
    aud: str
    exp: int
    iss: str
    sub: str
    cmp_id: str | None
    apps_id: str
    app_user_id: str
    issue_type: int


class UserSchema(BaseModel):
    swit_id: str
    access_token: str
    refresh_token: str

    class Config:
        orm_mode = True



import unittest

from switcore.ui.info_card import InfoCard
from switcore.ui.item import Item
from switcore.ui.text_paragraph import TextParagraph


class InfoCardTest(unittest.TestCase):

    def test_valid(self):
        text = TextParagraph(
            content="test content"
        )

        item = Item(
            label='test',
            text=text
        )

        info_card = InfoCard(
            action_id='test_action_id',
            draggable=True,
            items=[item])

        expected = {
            'type': 'info_card',
            'action_id': 'test_action_id',
            'draggable': True,
            'items': [
                {
                    'label': 'test',
                    'text': {
                        'content': 'test content',
                        'markdown': False,
                        'type': 'text'
                    }
                }
            ]
        }

        self.assertEqual(expected, info_card.dict(exclude_none=True))



from pydantic import BaseModel

from switcore.ui.select_item import SelectItem


class Tabs(BaseModel):
    """
        An element representing an array of tabs.
    """
    type: str = "tabs"
    tabs: list[SelectItem]
    value: str



import os

from starlette.requests import Request

from switcore.action.schemas import SwitRequest
from switcore.auth.exception import VerificationFailedException
from switcore.auth.signature_verification import SignatureVerifier


async def get_swit_request(
        request: Request,
):
    swit_singing_key: str = os.getenv('SWIT_SINGING_KEY', None)

    assert swit_singing_key is not None, "SWIT_SINGING_KEY is not set check .env file"

    verifier = SignatureVerifier(signing_key=swit_singing_key)

    timestamp_or_null: str | None = request.headers.get("x-swit-request-timestamp", None)
    signature_or_null: str | None = request.headers.get("x-swit-signature", None)

    body = await request.body()

    if not verifier.is_valid(body.decode(), timestamp_or_null, signature_or_null):
        raise VerificationFailedException("Verification failed check your signing key")

    data: dict = await request.json()
    return SwitRequest(**data)



from typing import Any, Optional, Dict

from fastapi import HTTPException
from starlette import status


class UndefinedSubmitAction(HTTPException):

    def __init__(self, detail: Any = None, headers: Optional[Dict[str, str]] = None) -> None:
        super().__init__(status_code=status.HTTP_404_NOT_FOUND, detail=detail, headers=headers)



from pydantic import BaseModel
from enum import Enum

from switcore.ui.element_components import StaticAction


class FileTypes(str, Enum):
    image = "image"
    video = "video"
    document = "document"
    pdf = "pdf"
    presentation = "presentation"
    spreadsheet = "spreadsheet"
    archive = "archive"
    psd = "psd"
    ai = "ai"
    other = "other"


def get_file_type(file_extension: str) -> FileTypes:
    if file_extension in ['jpg', 'jpeg', 'png', 'gif']:
        file_type = FileTypes.image
    elif file_extension in ['mp4', 'avi', 'mov']:
        file_type = FileTypes.video
    elif file_extension in ['doc', 'docx', 'txt']:
        file_type = FileTypes.document
    elif file_extension == 'pdf':
        file_type = FileTypes.pdf
    elif file_extension in ['ppt', 'pptx']:
        file_type = FileTypes.presentation
    elif file_extension == 'xls':
        file_type = FileTypes.spreadsheet
    elif file_extension in ['zip', 'rar']:
        file_type = FileTypes.archive
    elif file_extension == 'psd':
        file_type = FileTypes.psd
    elif file_extension == 'ai':
        file_type = FileTypes.ai
    else:
        file_type = FileTypes.other

    return file_type


class File(BaseModel):
    type: str = "file"
    file_type: FileTypes
    file_size: int
    file_name: str
    action_id: str | None
    static_action: StaticAction | None



from enum import Enum

from pydantic import BaseModel, validator


class TextParagraphColorTypes(str, Enum):
    gray900 = "gray900"
    gray800 = "gray800"
    gray700 = "gray700"


class TextParagraphSizeTypes(str, Enum):
    medium = "medium"
    large = "large"


class TextParagraphStyle(BaseModel):
    bold: bool = False
    color: TextParagraphColorTypes = TextParagraphColorTypes.gray900
    size: TextParagraphSizeTypes = TextParagraphSizeTypes.medium
    max_lines: int = 0

    @validator('max_lines', pre=True, always=True)
    def check_max_lines(cls, v):
        if v < 0 or v > 2:
            raise ValueError('max_lines must be between 0 and 2 inclusive.')
        return v


class TextParagraph(BaseModel):
    type: str = "text"
    markdown: bool = False
    content: str
    style: TextParagraphStyle | None = None



from typing import Any, Optional, Dict

from fastapi import HTTPException
from starlette import status


class NotFoundException(HTTPException):

    def __init__(self, detail: Any = None, headers: Optional[Dict[str, str]] = None) -> None:
        super().__init__(status.HTTP_404_NOT_FOUND, detail, headers)


class NeedScopeException(HTTPException):
    def __init__(self, detail: Any = None, headers: Optional[Dict[str, str]] = None) -> None:
        super().__init__(status.HTTP_404_NOT_FOUND, detail, headers)


class VerificationFailedException(HTTPException):
    def __init__(self, detail: Any = None, headers: Optional[Dict[str, str]] = None) -> None:
        super().__init__(status.HTTP_401_UNAUTHORIZED, detail, headers)



from abc import ABC, abstractmethod
from urllib.parse import urlencode

import httpx
from pydantic import BaseModel
from starlette.responses import Response

from switcore.auth.schemas import SwitToken


class OAuth2(BaseModel):
    client_id: str
    client_secret: str
    base_url: str
    authorize_endpoint: str
    access_token_endpoint: str
    refresh_token_endpoint: str
    bot_redirect_url: str
    user_redirect_url: str
    request_headers: dict[str, str] = {
        "content-type": "application/x-www-form-urlencoded",
        "Accept": "application/json",
    }

    def get_bot_authorization_url(self, state: str | None = None) -> str:
        params = {
            "response_type": "code",
            "client_id": self.client_id,
            "scope": "app:install",
            "redirect_uri": self.base_url + self.bot_redirect_url,
        }

        if state is not None:
            params["state"] = state

        return f"{self.authorize_endpoint}?{urlencode(params)}"

    def get_user_authorization_url(self, scope: str, state: str | None = None) -> str:
        params = {
            "response_type": "code",
            "client_id": self.client_id,
            "scope": scope,
            "redirect_uri": self.base_url + self.user_redirect_url,
        }

        if state is not None:
            params["state"] = state

        return f"{self.authorize_endpoint}?{urlencode(params)}"

    async def get_authorization_url(
            self,
            redirect_uri: str,
            scope: str,
            state: str | None = None,
            **extras_params,
    ) -> str:
        params = {
            "response_type": "code",
            "client_id": self.client_id,
            "redirect_uri": redirect_uri,
            "scope": scope,
        }

        if state is not None:
            params["state"] = state

        if extras_params is not None:
            params = {**params, **extras_params}

        return f"{self.authorize_endpoint}?{urlencode(params)}"

    async def get_access_token(self, code: str, redirect_uri: str) -> SwitToken:
        async with httpx.AsyncClient() as client:
            data = {
                "grant_type": "authorization_code",
                "code": code,
                "redirect_uri": redirect_uri,
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }

            response = await client.post(
                self.access_token_endpoint,
                data=data,
                headers=self.request_headers,
            )

            data: dict[str, str] = response.json()
            # response.raise_for_status()  # TODO handle error
            if response.status_code >= 400:
                raise ValueError(data)

            return SwitToken(**data)

    async def refresh_token(self, refresh_token: str) -> SwitToken:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.refresh_token_endpoint,
                data={
                    "grant_type": "refresh_token",
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                    "refresh_token": refresh_token,
                },
                headers=self.request_headers,
            )

            data: dict[str, str] = response.json()

            if response.status_code >= 400:
                raise ValueError(data)

            return SwitToken(**data)


class OAuth2AuthorizeCallbackABC(ABC):
    def __init__(self, client: OAuth2, redirect_url: str) -> None:
        self.client = client
        self.redirect_url = redirect_url

    @abstractmethod
    async def process_access_token(self, access_token: SwitToken, state: str | None) -> Response:
        pass

    async def _get_access_token(self, code: str) -> SwitToken:
        return await self.client.get_access_token(code, self.redirect_url)

    async def __call__(
            self,
            code: str,
            state: str | None,
            error: str | None = None,
    ) -> Response:
        access_token = await self._get_access_token(code)
        return await self.process_access_token(access_token, state)



from enum import Enum

from pydantic import BaseModel


class StyleColorTypes(str, Enum):
    primary = "primary"
    secondary = "secondary"
    danger = "danger"


class StyleShapeTypes(str, Enum):
    rectangular = "rectangular"
    rounded = "rounded"


class Style(BaseModel):
    color: StyleColorTypes
    shape: StyleShapeTypes


class Tag(BaseModel):
    type: str = "tag"
    content: str
    style: Style | None = None


class SubText(BaseModel):
    type: str = "subtext"
    content: str


class OpenOauthPopup(BaseModel):
    action_type: str = "open_oauth_popup"
    link_url: str


class OpenLink(BaseModel):
    action_type: str = "open_link"
    link_url: str


class CloseView(BaseModel):
    action_type: str = "close_view"


class ClipboardCopy(BaseModel):
    action_type: str = "clipboard_copy"
    content: str


StaticAction = OpenOauthPopup | OpenLink | CloseView | ClipboardCopy



import logging


def get_logger() -> logging.Logger:
    logger = logging.getLogger('[swit-logger]')
    logger.setLevel(logging.INFO)

    if not any(isinstance(handler, logging.StreamHandler) for handler in logger.handlers):
        console_handler = logging.StreamHandler()
        console_formatter = logging.Formatter(
            '%(asctime)s - [%(levelname)s] - %(name)s - (PID: %(process)d) - (File: %(filename)s -'
            ' Function: %(funcName)s - Line: %(lineno)d) - %(message)s')
        console_handler.setFormatter(console_formatter)
        logger.addHandler(console_handler)

    return logger



import unittest
from unittest.mock import patch
from urllib.parse import unquote

from fastapi import Depends
from httpx import Response
from starlette import status
from starlette.requests import Request
from starlette.responses import Response as StarletteResponse, HTMLResponse
from starlette.testclient import TestClient

from switcore.auth.dependencies import get_oauth
from switcore.auth.oauth2 import OAuth2AuthorizeCallbackABC, OAuth2
from switcore.auth.router import router as auth_router
from switcore.auth.schemas import SwitToken
from tests.utils import create_fastapi_app


class OAuth2Test(unittest.TestCase):
    def setUp(self) -> None:
        self.app = create_fastapi_app()
        self.app.include_router(auth_router, prefix="/auth")
        self.client = TestClient(self.app)
        super().setUp()

    def test_bot_authorization_url(self):
        response: Response = self.client.get("/auth/bot", follow_redirects=False)
        expected_url: str = "https://openapi.swit.io/oauth/authorize?" \
                            "response_type=code" \
                            "&client_id=test_client_id" \
                            "&scope=app:install" \
                            "&redirect_uri=test_base_url/auth/callback/bot"
        decoded_url: str = unquote(response.headers["location"])
        self.assertEqual(expected_url, decoded_url)

    def test_oauth2_callback_handler(self):
        expected = {
            "access_token": "eyJhbGciOiJIUzUxMiIsInR5cxxxxxxx.xxxxxxxx.xxxxxxxxxx",
            "expires_in": 604800,
            "refresh_token": "QIDQFWZZXFCXXXXX_XXXXX",
            "scope": "app:install",
            "token_type": "Bearer"
        }

        class BotAuthorizeCallback(OAuth2AuthorizeCallbackABC):
            async def process_access_token(
                    _self,  # noqa F841
                    swit_token: SwitToken,
                    state: str | None
            ) -> StarletteResponse:
                self.assertEqual(swit_token.access_token, expected["access_token"])
                self.assertEqual(swit_token.expires_in, expected["expires_in"])
                self.assertEqual(swit_token.refresh_token, expected["refresh_token"])
                self.assertEqual(swit_token.scope, expected["scope"])
                self.assertEqual(swit_token.token_type, expected["token_type"])
                return HTMLResponse(content="<script>window.close()</script>", status_code=status.HTTP_200_OK)

        @self.app.get("/auth/callback/bot")
        async def bot_oauth_callback(
                request: Request,  # noqa F841
                code: str,
                state: str | None = None,
                oauth: OAuth2 = Depends(get_oauth)
        ):
            callback = BotAuthorizeCallback(
                client=oauth,
                redirect_url="test_base_url/auth/callback/bot",
            )

            return await callback(
                code=code,
                state=state,
            )

        with patch('httpx.AsyncClient.post', return_value=Response(200, json=expected)) as mock_post:
            response = self.client.get("/auth/callback/bot?code=test_code&state=test_state")
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.content, b"<script>window.close()</script>")
            mock_post.assert_called_once()



from enum import Enum

from pydantic import BaseModel

from switcore.ui.element_components import Tag
from switcore.ui.image import Image
from switcore.ui.select_item import SelectItem


class Option(SelectItem):
    image: Image | None = None
    tags: list[Tag] | None = None


class OptionGroup(BaseModel):
    label: str
    options: list[Option]


class SelectStyleTypes(str, Enum):
    filled = "filled"
    outlined = "outlined"
    ghost = "ghost"


class Style(BaseModel):
    variant: SelectStyleTypes = SelectStyleTypes.outlined


class SelectSearch(BaseModel):
    disabled: bool = False
    placeholder: str | None = None
    value: str | None = None
    action_id: str


class Select(BaseModel):
    type: str = 'select'
    placeholder: str | None
    multiselect: bool = False
    trigger_on_input: bool = False
    value: list[str] | None
    options: list[Option] = []
    option_groups: list[OptionGroup] = []
    style: Style | None = None
    search: SelectSearch | None = None



from pydantic import BaseModel

from switcore.ui.button import Button


class ContextMenuItem(BaseModel):
    label: str
    action_id: str


class Header(BaseModel):
    title: str
    subtitle: str | None
    context_menu: list[ContextMenuItem] | None
    buttons: list[Button] | None



import datetime
import unittest

from switcore.action.activity_handler_abc import ActivityHandlerABC
from switcore.action.activity_router import ActivityRouter, PathResolver
from switcore.action.exceptions import UndefinedSubmitAction
from switcore.action.schemas import SwitRequest, PlatformTypes, UserInfo, UserPreferences, Context, UserAction, \
    UserActionType, View, Body, BaseState, SwitResponse, ViewCallbackType
from switcore.ui.header import Header


class ActivityHandler(ActivityHandlerABC):

    async def on_right_panel_open(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        pass

    async def on_presence_sync(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        pass

    async def on_user_commands_chat(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        pass

    async def on_user_commands_chat_extension(self, swit_request: SwitRequest, state: BaseState, *args,
                                              **kwargs) -> SwitResponse:
        pass

    async def on_user_commands_chat_commenting(self, swit_request: SwitRequest, state: BaseState, *args,
                                               **kwargs) -> SwitResponse:
        pass

    async def on_user_commands_context_menus_message(self, swit_request: SwitRequest, state: BaseState, *args,
                                                     **kwargs) -> SwitResponse:
        pass

    async def on_user_commands_context_menus_message_comment(self, swit_request: SwitRequest, state: BaseState, *args,
                                                             **kwargs) -> SwitResponse:
        pass

    async def on_view_actions_drop(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        pass

    async def on_view_actions_oauth_complete(self, swit_request: SwitRequest, state: BaseState, *args,
                                             **kwargs) -> SwitResponse:
        pass


def create_swit_request(view_id: str, action_id: str):
    return SwitRequest(
        platform=PlatformTypes.DESKTOP,
        time=datetime.datetime.now(),
        app_id="test_app_id",
        user_info=UserInfo(
            user_id="test_user_id",
            organization_id="test_organization_id"
        ),
        user_preferences=UserPreferences(
            language="ko",
            time_zone_offset="+0900",
            color_theme="light"
        ),
        context=Context(
            workspace_id="test_workspace_id",
            channel_id="test_channel_id"
        ),
        user_action=UserAction(
            type=UserActionType.view_actions_submit,
            id=action_id,
            slash_command="test_slash_command",
        ),
        current_view=View(
            view_id=view_id,
            state="state",
            header=Header(
                title="test_title",
            ),
            body=Body(),
        ))


class RouterTest(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self) -> None:
        self.activity_handler = ActivityHandler()
        return await super().asyncSetUp()

    async def test_router01(self):
        activity_router = ActivityRouter()
        action_id: str = str(PathResolver("test_action_id01"))
        swit_request: SwitRequest = create_swit_request("right_panel", action_id)

        @activity_router.register("test_action_id01")
        async def draw_webhook_create(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router)

        swit_response = await self.activity_handler.on_turn(swit_request, BaseState())
        self.assertTrue(isinstance(swit_response, SwitResponse))

    async def test_router02(self):
        activity_router = ActivityRouter()
        path_resolver = PathResolver("test_action_id01", ["a", 1])
        swit_request: SwitRequest = create_swit_request("right_panel", str(path_resolver))

        @activity_router.register("test_action_id01", ["right_panel"])
        async def draw_webhook_create(request: SwitRequest, state: BaseState, a: str, number: int):  # noqa F811
            self.assertEqual(a, "a")
            self.assertEqual(number, 1)
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router)

        swit_response = await self.activity_handler.on_turn(swit_request, BaseState())
        self.assertTrue(isinstance(swit_response, SwitResponse))

    async def test_router03(self):
        activity_router01 = ActivityRouter()
        activity_router02 = ActivityRouter()
        action_id_01: str = str(PathResolver("test_action_id01"))
        action_id_02: str = str(PathResolver("test_action_id02"))

        swit_request_01: SwitRequest = create_swit_request("right_panel01", action_id_01)
        swit_request_02: SwitRequest = create_swit_request("right_panel02", action_id_02)

        @activity_router01.register("test_action_id01", ["right_panel01"])
        async def draw_func_01(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        @activity_router02.register("test_action_id02", ["right_panel02"])
        async def draw_func_02(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router01)
        self.activity_handler.include_activity_router(activity_router02)

        swit_response_01 = await self.activity_handler.on_turn(swit_request_01, BaseState())
        swit_response_02 = await self.activity_handler.on_turn(swit_request_02, BaseState())

        self.assertTrue(isinstance(swit_response_01, SwitResponse))
        self.assertTrue(isinstance(swit_response_02, SwitResponse))

    async def test_router04(self):
        activity_router = ActivityRouter()
        action_id_01: str = str(PathResolver("test_action_id01"))
        action_id_02: str = str(PathResolver("test_action_id02"))
        action_id_03: str = str(PathResolver("test_action_id03", ["a", 1]))

        swit_request_01: SwitRequest = create_swit_request("right_panel", action_id_01)
        swit_request_02: SwitRequest = create_swit_request("right_panel", action_id_02)
        swit_request_03: SwitRequest = create_swit_request("right_panel", action_id_03)

        invalid_swit_request_01: SwitRequest = create_swit_request("modal", action_id_01)

        @activity_router.register("test_action_id01", ["right_panel"])
        async def draw_func_01(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        @activity_router.register("test_action_id02", ["right_panel"])
        async def draw_func_02(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        @activity_router.register("test_action_id03")
        async def draw_func_03(request: SwitRequest, state: BaseState, a: str, number: int):  # noqa F811
            self.assertEqual(a, "a")
            self.assertEqual(number, 1)
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router)

        swit_response_01 = await self.activity_handler.on_turn(swit_request_01, BaseState())
        swit_response_02 = await self.activity_handler.on_turn(swit_request_02, BaseState())
        swit_response_03 = await self.activity_handler.on_turn(swit_request_03, BaseState())

        self.assertTrue(isinstance(swit_response_01, SwitResponse))
        self.assertTrue(isinstance(swit_response_02, SwitResponse))
        self.assertTrue(isinstance(swit_response_03, SwitResponse))

        with self.assertRaises(UndefinedSubmitAction):
            await self.activity_handler.on_turn(invalid_swit_request_01, BaseState())

    async def test_router_with_slash(self):
        activity_router = ActivityRouter()
        swit_request = create_swit_request("right_panel", str(PathResolver("test_action_id01/aa/bb", ["a/a/b", 1])))

        @activity_router.register("test_action_id01/aa/bb", ["right_panel"])
        async def draw_webhook_create(request: SwitRequest, state: BaseState, a: str, number: int):  # noqa F811
            self.assertEqual(a, "a/a/b")
            self.assertEqual(number, 1)
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router)

        swit_response = await self.activity_handler.on_turn(swit_request, BaseState())
        self.assertTrue(isinstance(swit_response, SwitResponse))
